@@ -1,10 +1,64 @@
 use crate::assembler::Assembler;
-use crate::generator::{fmod, powi, powi_mod, Generator};
+use crate::generator::{fmod, powi, powi_mod, setup_call_binary, setup_call_unary, Generator};
 use crate::utils::align_stack;
+use crate::utils::{DataType, Reg};
 
 mod asm;
 use asm::{Amd, RoundingMode};
 
+macro_rules! binop {
+    ($self:ident, $sse:ident, $avx:ident, $simd:ident, $dst:expr, $s1: expr, $s2: expr, $com:ident) => {
+        $self.flush($dst);
+
+        match $self.family {
+            AmdFamily::AvxScalar => $self.amd.$avx(ϕ($dst), ϕ($s1), ϕ($s2)),
+            AmdFamily::AvxVector => $self.amd.$simd(ϕ($dst), ϕ($s1), ϕ($s2)),
+            AmdFamily::SSEScalar => {
+                let (x, y) = $self.shrink($dst, $s1, $s2, $com);
+                $self.amd.$sse(ϕ(x), ϕ(y));
+            }
+        }
+    };
+}
+
+macro_rules! select {
+    ($self:ident, $sse:ident, $avx:ident, $simd:ident, $x:expr, $y: expr, $z: expr, $w: expr) => {
+        match $self.family {
+            AmdFamily::AvxScalar => $self.amd.$avx($x, $y, $z, $w),
+            AmdFamily::AvxVector => $self.amd.$simd($x, $y, $z, $w),
+            AmdFamily::SSEScalar => $self.amd.$sse($x, $y, $z, $w),
+        }
+    };
+    ($self:ident, $sse:ident, $avx:ident, $simd:ident, $x:expr, $y: expr, $z: expr) => {
+        match $self.family {
+            AmdFamily::AvxScalar => $self.amd.$avx($x, $y, $z),
+            AmdFamily::AvxVector => $self.amd.$simd($x, $y, $z),
+            AmdFamily::SSEScalar => $self.amd.$sse($x, $y, $z),
+        }
+    };
+    ($self:ident, $sse:ident, $avx:ident, $simd:ident, $x:expr, $y: expr) => {
+        match $self.family {
+            AmdFamily::AvxScalar => $self.amd.$avx($x, $y),
+            AmdFamily::AvxVector => $self.amd.$simd($x, $y),
+            AmdFamily::SSEScalar => $self.amd.$sse($x, $y),
+        }
+    };
+}
+
+macro_rules! uniop {
+    ($self:ident, $sse:ident, $avx:ident, $simd:ident, $dst:expr, $s1: expr) => {
+        $self.flush($dst);
+        select!($self, $sse, $avx, $simd, ϕ($dst), ϕ($s1));
+    };
+}
+
+macro_rules! roundop {
+    ($self:ident, $dst:expr, $s1: expr, $mode: expr) => {
+        $self.flush($dst);
+        select!($self, roundsd, vroundsd, vroundpd, ϕ($dst), ϕ($s1), $mode);
+    };
+}
+
 pub enum AmdFamily {
     AvxScalar,
     AvxVector,
@@ -14,7 +68,6 @@ pub enum AmdFamily {
 pub struct AmdGenerator {
     amd: Amd,
     family: AmdFamily,
-    r0: Option<u32>,
     mask: u32,
 }
 
@@ -23,12 +76,27 @@ const STATES: u8 = Amd::R13;
 const IDX: u8 = Amd::R12;
 const PARAMS: u8 = Amd::RBX;
 
+/*
+ *  ϕ translates a logical register number (in Reg) to a physical
+ *  register number, according to the ABI.
+ */
+fn ϕ(r: Reg) -> u8 {
+    match r {
+        Reg::Ret => 0,
+        Reg::Temp => 1,
+        Reg::Left => 0,
+        Reg::Right => 1,
+        Reg::Gen(dst) => dst + 2,
+    }
+}
+
+const RET: u8 = 0;
+
 impl AmdGenerator {
     pub fn new(family: AmdFamily) -> AmdGenerator {
         AmdGenerator {
-            amd: Amd::new(),
+            amd: Amd::new(DataType::F64),
             family,
-            r0: None,
             mask: if cfg!(target_family = "windows") {
                 0x003f
             } else {
@@ -45,19 +113,19 @@ impl AmdGenerator {
         a and/or b. Therefore, cannot assume a and b are intact
         after calling this function.
     */
-    fn shrink(&mut self, dst: u8, a: u8, b: u8, commutative: bool) -> (u8, u8) {
-        if dst == a {
-            (dst, b)
-        } else if dst == b {
+    fn shrink(&mut self, dst: Reg, s1: Reg, s2: Reg, commutative: bool) -> (Reg, Reg) {
+        if dst == s1 {
+            (dst, s2)
+        } else if dst == s2 {
             // difficult case: dst == b, dst != a
             if !commutative {
-                self.fxchg(a, b);
+                self.fxchg(s1, s2);
             };
-            (dst, a)
+            (dst, s1)
         } else {
             // dst != a, dst != b, a ?= b
-            self.fmov(dst, a);
-            (dst, b)
+            self.fmov(dst, s1);
+            (dst, s2)
         }
     }
 
@@ -133,53 +201,42 @@ impl AmdGenerator {
         }
     }
 
-    fn flush(&mut self, dst: u8) {
-        if dst == 0 {
-            if let Some(idx) = self.r0 {
-                match self.family {
-                    AmdFamily::AvxScalar => self.amd.vmovsd_mem_xmm(Amd::RSP, (idx * 8) as i32, 0),
-                    AmdFamily::AvxVector => self.amd.vmovpd_mem_ymm(Amd::RSP, (idx * 32) as i32, 0),
-                    AmdFamily::SSEScalar => self.amd.movsd_mem_xmm(Amd::RSP, (idx * 8) as i32, 0),
-                };
-            };
+    fn flush(&mut self, dst: Reg) {
+        let reg = ϕ(dst);
+        let m = 1 << reg;
 
-            self.r0 = None;
-        } else {
-            let m = 1 << dst;
-
-            if self.mask & m == 0 {
-                // println!("saving reg {}", dst);
-                match self.family {
-                    AmdFamily::AvxScalar => {
-                        self.amd.vmovsd_mem_xmm(Amd::RSP, 8 * (dst as i32), dst)
-                    }
-                    AmdFamily::AvxVector => {
-                        self.amd.vmovpd_mem_ymm(Amd::RSP, 32 * (dst as i32), dst)
-                    }
-                    AmdFamily::SSEScalar => self.amd.movsd_mem_xmm(Amd::RSP, 8 * (dst as i32), dst),
-                };
-            }
-
-            self.mask |= m;
+        if self.mask & m == 0 {
+            select!(
+                self,
+                movsd_mem_xmm,
+                vmovsd_mem_xmm,
+                vmovpd_mem_ymm,
+                Amd::RSP,
+                (self.reg_size() as i32) * (reg as i32),
+                reg
+            );
         }
+
+        self.mask |= m;
     }
 
     fn restore_regs(&mut self) {
-        let last = self.first_shadow() + self.count_shadows();
+        // let last = self.first_shadow() + self.count_shadows();
+        let last = ϕ(Reg::Gen(self.count_shadows()));
 
-        for dst in last..16 {
-            let m = 1 << dst;
+        for reg in last..16 {
+            let m = 1 << reg;
 
             if self.mask & m != 0 {
-                match self.family {
-                    AmdFamily::AvxScalar => {
-                        self.amd.vmovsd_xmm_mem(dst, Amd::RSP, 8 * (dst as i32))
-                    }
-                    AmdFamily::AvxVector => {
-                        self.amd.vmovpd_ymm_mem(dst, Amd::RSP, 32 * (dst as i32))
-                    }
-                    AmdFamily::SSEScalar => self.amd.movsd_xmm_mem(dst, Amd::RSP, 8 * (dst as i32)),
-                }
+                select!(
+                    self,
+                    movsd_xmm_mem,
+                    vmovsd_xmm_mem,
+                    vmovpd_ymm_mem,
+                    reg,
+                    Amd::RSP,
+                    (self.reg_size() as i32) * (reg as i32)
+                );
             }
         }
     }
@@ -193,7 +250,7 @@ impl AmdGenerator {
             self.amd.mov_mem_reg(Amd::RSP, 0x08, MEM);
             self.amd.mov_mem_reg(Amd::RSP, 0x10, PARAMS);
             self.amd.mov_mem_reg(Amd::RSP, 0x18, IDX);
-            self.amd.mov_mem_reg(Amd::RSP, 0x20, Amd::R13);
+            self.amd.mov_mem_reg(Amd::RSP, 0x20, STATES);
         } else {
             self.amd.sub_rsp(32);
             self.amd.mov_mem_reg(Amd::RSP, 0x00, MEM);
@@ -217,45 +274,32 @@ impl AmdGenerator {
             self.amd.add_rsp(32);
         }
     }
-}
 
-macro_rules! binop {
-    ($self:ident, $sse:ident, $avx:ident, $simd:ident, $dst:expr, $a: expr, $b: expr, $com:ident) => {
-        $self.flush($dst);
+    #[cfg(target_family = "unix")]
+    fn chkstk(&mut self, size: u32) {
+        self.amd.sub_rsp(size);
+    }
 
-        match $self.family {
-            AmdFamily::AvxScalar => $self.amd.$avx($dst, $a, $b),
-            AmdFamily::AvxVector => $self.amd.$simd($dst, $a, $b),
-            AmdFamily::SSEScalar => {
-                let (x, y) = $self.shrink($dst, $a, $b, $com);
-                $self.amd.$sse(x, y);
-            }
+    #[cfg(target_family = "windows")]
+    fn chkstk(&mut self, mut size: u32) {
+        const PAGE_SIZE: u32 = 4096;
+
+        while size > PAGE_SIZE {
+            self.amd.sub_rsp(PAGE_SIZE);
+            self.amd.mov_reg_mem(Amd::RAX, Amd::RSP, 0);
+            size -= PAGE_SIZE;
         }
-    };
-}
 
-macro_rules! roundop {
-    ($self:ident, $dst:expr, $r: expr, $mode: expr) => {
-        $self.flush($dst);
-
-        match $self.family {
-            AmdFamily::AvxScalar => $self.amd.vroundsd($dst, $r, $mode),
-            AmdFamily::AvxVector => $self.amd.vroundpd($dst, $r, $mode),
-            AmdFamily::SSEScalar => $self.amd.roundsd($dst, $r, $mode),
-        }
-    };
+        self.amd.sub_rsp(size);
+    }
 }
 
 impl Generator for AmdGenerator {
-    fn first_shadow(&self) -> u8 {
-        2
-    }
-
     fn count_shadows(&self) -> u8 {
         if cfg!(target_family = "windows") {
-            4
+            4 // xmm2-xmm5
         } else {
-            14
+            14 // xmm2-xmm15
         }
     }
 
@@ -271,251 +315,263 @@ impl Generator for AmdGenerator {
     }
 
     fn three_address(&self) -> bool {
-        match self.family {
-            AmdFamily::AvxScalar => true,
-            AmdFamily::AvxVector => true,
-            AmdFamily::SSEScalar => false,
-        }
+        !matches!(self.family, AmdFamily::SSEScalar)
     }
 
     //***********************************
-    fn fmov(&mut self, dst: u8, r: u8) {
-        if dst == r {
-            return;
-        }
 
-        self.flush(dst);
-
-        match self.family {
-            AmdFamily::AvxScalar | AmdFamily::AvxVector => self.amd.vmovapd(dst, r),
-            AmdFamily::SSEScalar => self.amd.movapd(dst, r),
+    fn fmov(&mut self, dst: Reg, s1: Reg) {
+        if dst != s1 {
+            self.flush(dst);
+            select!(self, movapd, vmovapd, vmovapd, ϕ(dst), ϕ(s1));
         }
     }
 
-    fn fxchg(&mut self, a: u8, b: u8) {
-        self.flush(a);
-        self.flush(b);
+    fn fxchg(&mut self, s1: Reg, s2: Reg) {
+        self.flush(s1);
+        self.flush(s2);
 
         match self.family {
             AmdFamily::AvxScalar | AmdFamily::AvxVector => {
-                self.amd.vxorpd(a, a, b);
-                self.amd.vxorpd(b, a, b);
-                self.amd.vxorpd(a, a, b);
+                self.amd.vxorpd(ϕ(s1), ϕ(s1), ϕ(s2));
+                self.amd.vxorpd(ϕ(s2), ϕ(s1), ϕ(s2));
+                self.amd.vxorpd(ϕ(s1), ϕ(s1), ϕ(s2));
             }
             AmdFamily::SSEScalar => {
-                self.amd.xorpd(a, b);
-                self.amd.xorpd(b, a);
-                self.amd.xorpd(a, b);
+                self.amd.xorpd(ϕ(s1), ϕ(s2));
+                self.amd.xorpd(ϕ(s2), ϕ(s1));
+                self.amd.xorpd(ϕ(s1), ϕ(s2));
             }
         }
     }
 
-    fn load_const(&mut self, dst: u8, label: &str) {
+    fn load_const(&mut self, dst: Reg, label: &str) {
         self.flush(dst);
-
-        match self.family {
-            AmdFamily::AvxScalar => self.amd.vmovsd_xmm_label(dst, label),
-            AmdFamily::AvxVector => self.amd.vbroadcastsd_label(dst, label),
-            AmdFamily::SSEScalar => self.amd.movsd_xmm_label(dst, label),
-        }
+        select!(
+            self,
+            movsd_xmm_label,
+            vmovsd_xmm_label,
+            vbroadcastsd_label,
+            ϕ(dst),
+            label
+        );
     }
 
-    fn load_mem(&mut self, dst: u8, idx: u32) {
+    fn load_mem(&mut self, dst: Reg, idx: u32) {
         self.flush(dst);
-
-        match self.family {
-            AmdFamily::AvxScalar => self.amd.vmovsd_xmm_mem(dst, MEM, (idx * 8) as i32),
-            AmdFamily::AvxVector => self.amd.vmovpd_ymm_mem(dst, MEM, (idx * 32) as i32),
-            AmdFamily::SSEScalar => self.amd.movsd_xmm_mem(dst, MEM, (idx * 8) as i32),
-        }
+        select!(
+            self,
+            movsd_xmm_mem,
+            vmovsd_xmm_mem,
+            vmovpd_ymm_mem,
+            ϕ(dst),
+            MEM,
+            (idx * self.reg_size()) as i32
+        );
     }
 
-    fn save_mem(&mut self, src: u8, idx: u32) {
-        match self.family {
-            AmdFamily::AvxScalar => self.amd.vmovsd_mem_xmm(MEM, (idx * 8) as i32, src),
-            AmdFamily::AvxVector => self.amd.vmovpd_mem_ymm(MEM, (idx * 32) as i32, src),
-            AmdFamily::SSEScalar => self.amd.movsd_mem_xmm(MEM, (idx * 8) as i32, src),
-        }
+    fn save_mem(&mut self, dst: Reg, idx: u32) {
+        select!(
+            self,
+            movsd_mem_xmm,
+            vmovsd_mem_xmm,
+            vmovpd_mem_ymm,
+            MEM,
+            (idx * self.reg_size()) as i32,
+            ϕ(dst)
+        );
     }
 
-    fn load_param(&mut self, dst: u8, idx: u32) {
+    fn save_mem_result(&mut self, idx: u32) {
+        self.save_mem(Reg::Ret, idx);
+    }
+
+    fn load_param(&mut self, dst: Reg, idx: u32) {
         self.flush(dst);
-
-        match self.family {
-            AmdFamily::AvxScalar => self.amd.vmovsd_xmm_mem(dst, PARAMS, (idx * 8) as i32),
-            AmdFamily::AvxVector => self.amd.vbroadcastsd(dst, PARAMS, (idx * 8) as i32),
-            AmdFamily::SSEScalar => self.amd.movsd_xmm_mem(dst, PARAMS, (idx * 8) as i32),
-        }
+        select!(
+            self,
+            movsd_xmm_mem,
+            vmovsd_xmm_mem,
+            vbroadcastsd,
+            ϕ(dst),
+            PARAMS,
+            8 * idx as i32
+        );
     }
 
-    fn load_stack(&mut self, dst: u8, idx: u32) {
-        if let Some(k) = self.r0 {
-            if k == idx {
-                match self.family {
-                    AmdFamily::AvxScalar | AmdFamily::AvxVector => self.amd.vmovapd(dst, 0),
-                    AmdFamily::SSEScalar => {}
-                };
-                self.r0 = None;
-                return;
-            }
-        };
-
-        match self.family {
-            AmdFamily::AvxScalar => self.amd.vmovsd_xmm_mem(dst, Amd::RSP, (idx * 8) as i32),
-            AmdFamily::AvxVector => self.amd.vmovpd_ymm_mem(dst, Amd::RSP, (idx * 32) as i32),
-            AmdFamily::SSEScalar => self.amd.movsd_xmm_mem(dst, Amd::RSP, (idx * 8) as i32),
-        }
-    }
-
-    fn save_stack(&mut self, src: u8, idx: u32) {
-        if src == 0 && !matches!(self.family, AmdFamily::SSEScalar) {
-            self.r0 = Some(idx);
-            return;
-        }
-
-        match self.family {
-            AmdFamily::AvxScalar => self.amd.vmovsd_mem_xmm(Amd::RSP, (idx * 8) as i32, src),
-            AmdFamily::AvxVector => self.amd.vmovpd_mem_ymm(Amd::RSP, (idx * 32) as i32, src),
-            AmdFamily::SSEScalar => self.amd.movsd_mem_xmm(Amd::RSP, (idx * 8) as i32, src),
-        }
-    }
-
-    fn neg(&mut self, dst: u8, r: u8) {
+    fn load_stack(&mut self, dst: Reg, idx: u32) {
         self.flush(dst);
-        self.load_const(1, "_minus_zero_");
-        self.xor(dst, r, 1);
+        select!(
+            self,
+            movsd_xmm_mem,
+            vmovsd_xmm_mem,
+            vmovpd_ymm_mem,
+            ϕ(dst),
+            Amd::RSP,
+            (idx * self.reg_size()) as i32
+        );
     }
 
-    fn abs(&mut self, dst: u8, r: u8) {
+    fn save_stack(&mut self, dst: Reg, idx: u32) {
+        select!(
+            self,
+            movsd_mem_xmm,
+            vmovsd_mem_xmm,
+            vmovpd_mem_ymm,
+            Amd::RSP,
+            (idx * self.reg_size()) as i32,
+            ϕ(dst)
+        );
+    }
+
+    fn save_stack_result(&mut self, idx: u32) {
+        self.save_stack(Reg::Ret, idx);
+    }
+
+    fn neg(&mut self, dst: Reg, s1: Reg) {
         self.flush(dst);
-        self.load_const(1, "_minus_zero_");
-        self.andnot(dst, 1, r);
+        self.load_const(Reg::Temp, "_minus_zero_");
+        self.xor(dst, s1, Reg::Temp);
     }
 
-    fn root(&mut self, dst: u8, r: u8) {
+    fn abs(&mut self, dst: Reg, s1: Reg) {
         self.flush(dst);
-        match self.family {
-            AmdFamily::AvxScalar => self.amd.vsqrtsd(dst, r),
-            AmdFamily::AvxVector => self.amd.vsqrtpd(dst, r),
-            AmdFamily::SSEScalar => self.amd.sqrtsd(dst, r),
-        }
+        self.load_const(Reg::Temp, "_minus_zero_");
+        self.andnot(dst, Reg::Temp, s1);
     }
 
-    fn square(&mut self, dst: u8, r: u8) {
+    fn root(&mut self, dst: Reg, s1: Reg) {
+        uniop!(self, sqrtsd, vsqrtsd, vsqrtpd, dst, s1);
+    }
+
+    fn square(&mut self, dst: Reg, s1: Reg) {
         self.flush(dst);
-        self.times(dst, r, r);
+        self.times(dst, s1, s1);
     }
 
-    fn cube(&mut self, dst: u8, r: u8) {
+    fn cube(&mut self, dst: Reg, s1: Reg) {
         self.flush(dst);
-        self.times(1, r, r);
-        self.times(dst, r, 1);
+        self.times(Reg::Temp, s1, s1);
+        self.times(dst, s1, Reg::Temp);
     }
 
-    fn powi(&mut self, dst: u8, r: u8, power: i32) {
+    fn powi(&mut self, dst: Reg, s1: Reg, power: i32) {
         self.flush(dst);
         if power == 0 {
             self.load_const(dst, "_one_");
         } else {
-            powi(self, dst, r, power);
+            powi(self, dst, s1, power);
         }
     }
 
-    fn powi_mod(&mut self, dst: u8, r: u8, power: i32, modulus: u8) {
+    fn powi_mod(&mut self, dst: Reg, s1: Reg, power: i32, modulus: Reg) {
         self.flush(dst);
         if power == 0 {
             self.load_const(dst, "_one_");
         } else {
-            powi_mod(self, dst, r, power, modulus);
+            powi_mod(self, dst, s1, power, modulus);
         }
     }
 
-    fn recip(&mut self, dst: u8, r: u8) {
+    fn recip(&mut self, dst: Reg, s1: Reg) {
         self.flush(dst);
-        self.load_const(1, "_one_");
-        self.divide(dst, 1, r);
+        self.load_const(Reg::Temp, "_one_");
+        self.divide(dst, Reg::Temp, s1);
     }
 
-    fn round(&mut self, dst: u8, r: u8) {
-        roundop!(self, dst, r, RoundingMode::Round);
+    fn round(&mut self, dst: Reg, s1: Reg) {
+        roundop!(self, dst, s1, RoundingMode::Round);
     }
 
-    fn floor(&mut self, dst: u8, r: u8) {
-        roundop!(self, dst, r, RoundingMode::Floor);
+    fn floor(&mut self, dst: Reg, s1: Reg) {
+        roundop!(self, dst, s1, RoundingMode::Floor);
     }
 
-    fn ceiling(&mut self, dst: u8, r: u8) {
-        roundop!(self, dst, r, RoundingMode::Ceiling);
+    fn ceiling(&mut self, dst: Reg, s1: Reg) {
+        roundop!(self, dst, s1, RoundingMode::Ceiling);
     }
 
-    fn trunc(&mut self, dst: u8, r: u8) {
-        roundop!(self, dst, r, RoundingMode::Trunc);
+    fn trunc(&mut self, dst: Reg, s1: Reg) {
+        roundop!(self, dst, s1, RoundingMode::Trunc);
     }
 
-    fn fmod(&mut self, dst: u8, a: u8, b: u8) {
-        fmod(self, dst, a, b);
+    fn frac(&mut self, dst: Reg, s1: Reg) {
+        self.floor(Reg::Temp, s1);
+        self.minus(dst, s1, Reg::Temp);
     }
 
-    fn plus(&mut self, dst: u8, a: u8, b: u8) {
-        binop!(self, addsd, vaddsd, vaddpd, dst, a, b, true);
+    fn fmod(&mut self, dst: Reg, s1: Reg, s2: Reg) {
+        fmod(self, dst, s1, s2);
     }
 
-    fn minus(&mut self, dst: u8, a: u8, b: u8) {
-        binop!(self, subsd, vsubsd, vsubpd, dst, a, b, false);
+    fn plus(&mut self, dst: Reg, s1: Reg, s2: Reg) {
+        binop!(self, addsd, vaddsd, vaddpd, dst, s1, s2, true);
     }
 
-    fn times(&mut self, dst: u8, a: u8, b: u8) {
-        binop!(self, mulsd, vmulsd, vmulpd, dst, a, b, true);
+    fn minus(&mut self, dst: Reg, s1: Reg, s2: Reg) {
+        binop!(self, subsd, vsubsd, vsubpd, dst, s1, s2, false);
     }
 
-    fn divide(&mut self, dst: u8, a: u8, b: u8) {
-        binop!(self, divsd, vdivsd, vdivpd, dst, a, b, false);
+    fn times(&mut self, dst: Reg, s1: Reg, s2: Reg) {
+        binop!(self, mulsd, vmulsd, vmulpd, dst, s1, s2, true);
     }
 
-    fn gt(&mut self, dst: u8, a: u8, b: u8) {
-        binop!(self, cmpnlesd, vcmpnlesd, vcmpnlepd, dst, a, b, false);
+    fn divide(&mut self, dst: Reg, s1: Reg, s2: Reg) {
+        binop!(self, divsd, vdivsd, vdivpd, dst, s1, s2, false);
     }
 
-    fn geq(&mut self, dst: u8, a: u8, b: u8) {
-        binop!(self, cmpnltsd, vcmpnltsd, vcmpnltpd, dst, a, b, false);
+    fn gt(&mut self, dst: Reg, s1: Reg, s2: Reg) {
+        binop!(self, cmpnlesd, vcmpnlesd, vcmpnlepd, dst, s1, s2, false);
     }
 
-    fn lt(&mut self, dst: u8, a: u8, b: u8) {
-        binop!(self, cmpltsd, vcmpltsd, vcmpltpd, dst, a, b, false);
+    fn geq(&mut self, dst: Reg, s1: Reg, s2: Reg) {
+        binop!(self, cmpnltsd, vcmpnltsd, vcmpnltpd, dst, s1, s2, false);
     }
 
-    fn leq(&mut self, dst: u8, a: u8, b: u8) {
-        binop!(self, cmplesd, vcmplesd, vcmplepd, dst, a, b, false);
+    fn lt(&mut self, dst: Reg, s1: Reg, s2: Reg) {
+        binop!(self, cmpltsd, vcmpltsd, vcmpltpd, dst, s1, s2, false);
     }
 
-    fn eq(&mut self, dst: u8, a: u8, b: u8) {
-        binop!(self, cmpeqsd, vcmpeqsd, vcmpeqpd, dst, a, b, true);
+    fn leq(&mut self, dst: Reg, s1: Reg, s2: Reg) {
+        binop!(self, cmplesd, vcmplesd, vcmplepd, dst, s1, s2, false);
     }
 
-    fn neq(&mut self, dst: u8, a: u8, b: u8) {
-        binop!(self, cmpneqsd, vcmpneqsd, vcmpneqpd, dst, a, b, true);
+    fn eq(&mut self, dst: Reg, s1: Reg, s2: Reg) {
+        binop!(self, cmpeqsd, vcmpeqsd, vcmpeqpd, dst, s1, s2, true);
     }
 
-    fn and(&mut self, dst: u8, a: u8, b: u8) {
-        binop!(self, andpd, vandpd, vandpd, dst, a, b, true);
+    fn neq(&mut self, dst: Reg, s1: Reg, s2: Reg) {
+        binop!(self, cmpneqsd, vcmpneqsd, vcmpneqpd, dst, s1, s2, true);
     }
 
-    fn andnot(&mut self, dst: u8, a: u8, b: u8) {
-        binop!(self, andnpd, vandnpd, vandnpd, dst, a, b, false);
+    fn and(&mut self, dst: Reg, s1: Reg, s2: Reg) {
+        binop!(self, andpd, vandpd, vandpd, dst, s1, s2, true);
     }
 
-    fn or(&mut self, dst: u8, a: u8, b: u8) {
-        binop!(self, orpd, vorpd, vorpd, dst, a, b, true);
+    fn andnot(&mut self, dst: Reg, s1: Reg, s2: Reg) {
+        binop!(self, andnpd, vandnpd, vandnpd, dst, s1, s2, false);
     }
 
-    fn xor(&mut self, dst: u8, a: u8, b: u8) {
-        binop!(self, xorpd, vxorpd, vxorpd, dst, a, b, true);
+    fn or(&mut self, dst: Reg, s1: Reg, s2: Reg) {
+        binop!(self, orpd, vorpd, vorpd, dst, s1, s2, true);
     }
 
-    fn not(&mut self, dst: u8, r: u8) {
+    fn xor(&mut self, dst: Reg, s1: Reg, s2: Reg) {
+        binop!(self, xorpd, vxorpd, vxorpd, dst, s1, s2, true);
+    }
+
+    fn not(&mut self, dst: Reg, s1: Reg) {
         self.flush(dst);
-        self.load_const(1, "_all_ones_");
-        self.xor(dst, r, 1);
+        self.load_const(Reg::Temp, "_all_ones_");
+        self.xor(dst, s1, Reg::Temp);
+    }
+
+    fn setup_call_unary(&mut self, s1: Reg) {
+        setup_call_unary(self, s1);
+    }
+
+    fn setup_call_binary(&mut self, s1: Reg, s2: Reg) {
+        setup_call_binary(self, s1, s2);
     }
 
     fn call(&mut self, label: &str, num_args: usize) {
@@ -558,14 +614,14 @@ impl Generator for AmdGenerator {
             self.amd.jmp(false_label);
         }
     */
-    fn select_if(&mut self, dst: u8, cond: u8, a: u8) {
+    fn select_if(&mut self, dst: Reg, cond: Reg, s1: Reg) {
         self.flush(dst);
-        self.amd.vandpd(dst, cond, a);
+        self.amd.vandpd(ϕ(dst), ϕ(cond), ϕ(s1));
     }
 
-    fn select_else(&mut self, dst: u8, cond: u8, a: u8) {
+    fn select_else(&mut self, dst: Reg, cond: Reg, s1: Reg) {
         self.flush(dst);
-        self.amd.vandnpd(dst, cond, a);
+        self.amd.vandnpd(ϕ(dst), ϕ(cond), ϕ(s1));
     }
 
     /****************** Prologues/Epilogues ********************/
@@ -639,33 +695,39 @@ impl Generator for AmdGenerator {
         self.amd.jz("@main");
 
         let size = (count_states + count_obs + 1) as u32 * self.reg_size();
-        self.amd.sub_rsp(size);
+        //self.amd.sub_rsp(size);
+        self.chkstk(size);
         self.amd.mov(MEM, Amd::RSP); // in indirect mode, MEM is allocated on the stack
 
         for i in 0..count_states {
             self.amd.mov_reg_mem(Amd::RAX, STATES, 8 * i as i32);
             let k = i as u32 * self.reg_size();
-
-            match self.family {
-                AmdFamily::AvxScalar => {
-                    self.amd.vmovsd_xmm_indexed(0, Amd::RAX, IDX, 8);
-                    self.amd.vmovsd_mem_xmm(MEM, k as i32, 0);
-                }
-                AmdFamily::AvxVector => {
-                    self.amd.vmovpd_ymm_indexed(0, Amd::RAX, IDX, 8);
-                    self.amd.vmovpd_mem_ymm(MEM, k as i32, 0);
-                }
-                AmdFamily::SSEScalar => {
-                    self.amd.movsd_xmm_indexed(0, Amd::RAX, IDX, 8);
-                    self.amd.movsd_mem_xmm(MEM, k as i32, 0);
-                }
-            }
+            select!(
+                self,
+                movsd_xmm_indexed,
+                vmovsd_xmm_indexed,
+                vmovpd_ymm_indexed,
+                RET,
+                Amd::RAX,
+                IDX,
+                8
+            );
+            select!(
+                self,
+                movsd_mem_xmm,
+                vmovsd_mem_xmm,
+                vmovpd_mem_ymm,
+                MEM,
+                k as i32,
+                RET
+            );
         }
 
         // may save idx (RDX) as double in RBP + 8/32 * count_states
 
         self.set_label("@main");
-        self.amd.sub_rsp(self.frame_size(cap));
+        // self.amd.sub_rsp(self.frame_size(cap));
+        self.chkstk(self.frame_size(cap));
     }
 
     fn epilogue_indirect(&mut self, cap: u32, count_states: usize, count_obs: usize) {
@@ -682,21 +744,25 @@ impl Generator for AmdGenerator {
             self.amd
                 .mov_reg_mem(Amd::RAX, STATES, 8 * (count_states + i) as i32);
             let k = (count_states + i + 1) as u32 * self.reg_size();
-
-            match self.family {
-                AmdFamily::AvxScalar => {
-                    self.amd.vmovsd_xmm_mem(0, MEM, k as i32);
-                    self.amd.vmovsd_indexed_xmm(Amd::RAX, IDX, 8, 0);
-                }
-                AmdFamily::AvxVector => {
-                    self.amd.vmovpd_ymm_mem(0, MEM, k as i32);
-                    self.amd.vmovpd_indexed_ymm(Amd::RAX, IDX, 8, 0);
-                }
-                AmdFamily::SSEScalar => {
-                    self.amd.movsd_xmm_mem(0, MEM, k as i32);
-                    self.amd.movsd_indexed_xmm(Amd::RAX, IDX, 8, 0);
-                }
-            }
+            select!(
+                self,
+                movsd_xmm_mem,
+                vmovsd_xmm_mem,
+                vmovpd_ymm_mem,
+                RET,
+                MEM,
+                k as i32
+            );
+            select!(
+                self,
+                movsd_indexed_xmm,
+                vmovsd_indexed_xmm,
+                vmovpd_indexed_ymm,
+                Amd::RAX,
+                IDX,
+                8,
+                RET
+            );
         }
 
         self.amd.add_rsp(size);
