@@ -1,11 +1,15 @@
 use anyhow::{anyhow, Result};
 
 use std::cell::RefCell;
+use std::fmt;
+use std::fmt::Debug;
+use std::hash::{DefaultHasher, Hash, Hasher};
 use std::rc::Rc;
 
 use crate::generator::Generator;
 use crate::symbol::{Loc, Symbol};
-use crate::utils::{bool_to_f64, Eval};
+use crate::utils::{bool_to_f64, reg, Eval};
+use crate::COUNT_SCRATCH;
 
 #[derive(Debug, Clone)]
 pub enum VarStatus {
@@ -16,7 +20,9 @@ pub enum VarStatus {
     Singular,
 }
 
-#[derive(Debug, Clone)]
+const COMMUTATIVE: &[&str] = &["plus", "times", "eq", "neq", "and", "or", "xor"];
+
+#[derive(Clone)]
 pub enum Node {
     Void,
     Const {
@@ -30,19 +36,56 @@ pub enum Node {
     Unary {
         op: String,
         arg: Box<Node>,
-        ershov: u8,
         power: i32,
+        ershov: u8,
+        h: u64,
+        w: u32,
     },
     Binary {
         op: String,
         left: Box<Node>,
         right: Box<Node>,
-        ershov: u8,
         power: i32,
+        ershov: u8,
+        h: u64,
+        w: u32,
     },
 }
 
 impl Node {
+    pub fn hashof(&self) -> u64 {
+        let mut hasher = DefaultHasher::new();
+
+        match self {
+            Node::Void => b"void".hash(&mut hasher),
+            Node::Const { idx, .. } => {
+                b"const".hash(&mut hasher);
+                idx.hash(&mut hasher);
+            }
+            Node::Var { sym, .. } => {
+                b"var".hash(&mut hasher);
+                sym.borrow().hash(&mut hasher);
+            }
+            Node::Unary { h, .. } => {
+                return *h;
+            }
+            Node::Binary { h, .. } => {
+                return *h;
+            }
+        };
+
+        hasher.finish()
+    }
+
+    pub fn weightof(&self) -> u32 {
+        match self {
+            Node::Void => 0,
+            Node::Const { .. } | Node::Var { .. } => 1,
+            Node::Unary { w, .. } => *w,
+            Node::Binary { w, .. } => *w,
+        }
+    }
+
     pub fn create_void() -> Node {
         Node::Void
     }
@@ -58,41 +101,95 @@ impl Node {
         }
     }
 
-    pub fn create_unary(op: &str, arg: Node) -> Node {
+    pub fn create_unary(op: &str, arg: Node, power: i32) -> Node {
+        let e = arg.ershov_number();
+
+        let mut hasher = DefaultHasher::new();
+        op.hash(&mut hasher);
+        arg.hashof().hash(&mut hasher);
+        power.hash(&mut hasher);
+
+        let w = 1 + arg.weightof();
+
         Node::Unary {
             op: op.to_string(),
             arg: Box::new(arg),
-            ershov: 0,
-            power: 1,
+            ershov: e,
+            power,
+            h: hasher.finish(),
+            w,
         }
     }
 
-    pub fn create_binary(op: &str, left: Node, right: Node) -> Node {
+    pub fn create_binary(op: &str, left: Node, right: Node, power: i32) -> Node {
+        let e = Self::calc_ershov(&left, &right);
+
+        let mut hasher = DefaultHasher::new();
+        op.hash(&mut hasher);
+        power.hash(&mut hasher);
+
+        let mut l = left.hashof();
+        let mut r = right.hashof();
+
+        (l, r) = if COMMUTATIVE.contains(&op) && l > r {
+            (r, l)
+        } else {
+            (l, r)
+        };
+
+        l.hash(&mut hasher);
+        r.hash(&mut hasher);
+
+        let w = 1 + left.weightof() + right.weightof();
+
         Node::Binary {
             op: op.to_string(),
             left: Box::new(left),
             right: Box::new(right),
-            ershov: 0,
-            power: 1,
+            ershov: e,
+            power,
+            h: hasher.finish(),
+            w,
         }
     }
 
     pub fn create_powi(arg: Node, power: i32) -> Node {
-        Node::Unary {
-            op: "_powi_".to_string(),
-            arg: Box::new(arg),
-            ershov: 0,
-            power,
-        }
+        Self::create_unary("_powi_", arg, power)
     }
 
     pub fn create_modular_powi(left: Node, right: Node, power: i32) -> Node {
-        Node::Binary {
-            op: "_powi_mod_".to_string(),
-            left: Box::new(left),
-            right: Box::new(right),
-            ershov: 0,
-            power,
+        Self::create_binary("_powi_mod_", left, right, power)
+    }
+
+    pub fn first(&mut self) -> Option<&mut Node> {
+        match self {
+            Node::Unary { arg, .. } => Some(arg),
+            Node::Binary { left, right, .. } => {
+                let el = left.ershov_number();
+                let er = right.ershov_number();
+                if el >= er {
+                    Some(left)
+                } else {
+                    Some(right)
+                }
+            }
+            _ => None,
+        }
+    }
+
+    pub fn second(&mut self) -> Option<&mut Node> {
+        match self {
+            Node::Unary { .. } => None,
+            Node::Binary { left, right, .. } => {
+                let el = left.ershov_number();
+                let er = right.ershov_number();
+                if el >= er {
+                    Some(right)
+                } else {
+                    Some(left)
+                }
+            }
+            _ => None,
         }
     }
 
@@ -102,29 +199,15 @@ impl Node {
     /// Note the twist in the middle. The decision to traverse left
     /// or right link depends on ershov number of each link.
     fn postorder_forward(&mut self, f: fn(&mut Node)) {
-        match self {
-            Node::Void | Node::Const { .. } | Node::Var { .. } => {
-                f(self);
-            }
-            Node::Unary { arg, .. } => {
-                arg.postorder_forward(f);
-                f(self);
-            }
-            Node::Binary { left, right, .. } => {
-                let el = left.ershov_number();
-                let er = right.ershov_number();
+        if let Some(n) = self.first() {
+            n.postorder_forward(f)
+        };
 
-                if el >= er {
-                    left.postorder_forward(f);
-                    right.postorder_forward(f);
-                } else {
-                    right.postorder_forward(f);
-                    left.postorder_forward(f);
-                };
+        if let Some(n) = self.second() {
+            n.postorder_forward(f)
+        };
 
-                f(self);
-            }
-        }
+        f(self);
     }
 
     /// postorder_backward does a backward postorder traversal of the
@@ -133,29 +216,15 @@ impl Node {
     /// Note the twist in the middle. The decision to traverse left
     /// or right link depends on ershov number of each link.
     fn postorder_backward(&mut self, f: fn(&mut Node)) {
-        match self {
-            Node::Void | Node::Const { .. } | Node::Var { .. } => {
-                f(self);
-            }
-            Node::Unary { arg, .. } => {
-                arg.postorder_backward(f);
-                f(self);
-            }
-            Node::Binary { left, right, .. } => {
-                let el = left.ershov_number();
-                let er = right.ershov_number();
+        if let Some(n) = self.second() {
+            n.postorder_forward(f)
+        };
 
-                if el >= er {
-                    right.postorder_backward(f);
-                    left.postorder_backward(f);
-                } else {
-                    left.postorder_backward(f);
-                    right.postorder_backward(f);
-                };
+        if let Some(n) = self.first() {
+            n.postorder_forward(f)
+        };
 
-                f(self);
-            }
-        }
+        f(self);
     }
 
     /// Ershov number is the number of temporary registers needed to
@@ -168,23 +237,14 @@ impl Node {
         }
     }
 
-    fn ershov_func(&mut self) {
-        match self {
-            Node::Unary { arg, ershov, .. } => {
-                *ershov = arg.ershov_number();
-            }
-            Node::Binary {
-                left,
-                right,
-                ershov,
-                ..
-            } => {
-                let l = left.ershov_number();
-                let r = right.ershov_number();
-                let e = if l == r { l + 1 } else { l.max(r) };
-                *ershov = e;
-            }
-            _ => {}
+    pub fn calc_ershov(left: &Node, right: &Node) -> u8 {
+        let l = left.ershov_number();
+        let r = right.ershov_number();
+
+        if l == r {
+            l + 1
+        } else {
+            l.max(r)
         }
     }
 
@@ -220,11 +280,10 @@ impl Node {
     /// The main entry point to compile an expression tree
     /// should be called on the root of the expression tree
     pub fn compile_tree(&mut self, ir: &mut dyn Generator) -> Result<u8> {
-        self.postorder_forward(Self::ershov_func);
         self.postorder_forward(Self::mark_first);
         self.postorder_backward(Self::mark_last);
 
-        let last = ir.first_shadow() + self.ershov_number();
+        let last = self.ershov_number();
 
         // we check ir.three_address() because AmdGenerator::shrink may swap
         // registers when generating code for SSE (two-address code).
@@ -232,7 +291,7 @@ impl Node {
         // correctness first.
 
         let mut pool: Vec<u8> = if ir.three_address() {
-            (last..16).rev().collect()
+            (last..COUNT_SCRATCH).rev().collect()
         } else {
             Vec::new()
         };
@@ -253,10 +312,9 @@ impl Node {
 
     fn compile_const(&self, ir: &mut dyn Generator, base: u8) -> Result<u8> {
         if let Node::Const { idx, .. } = &self {
-            let r = ir.first_shadow() + base;
             let label = format!("_const_{}_", idx);
-            ir.load_const(r, &label);
-            Ok(r)
+            ir.load_const(reg(base), &label);
+            Ok(base)
         } else {
             unreachable!();
         }
@@ -264,9 +322,9 @@ impl Node {
 
     fn load_var(ir: &mut dyn Generator, dst: u8, loc: &Loc) -> u8 {
         match loc {
-            Loc::Stack(idx) => ir.load_stack(dst, *idx),
-            Loc::Mem(idx) => ir.load_mem(dst, *idx),
-            Loc::Param(idx) => ir.load_param(dst, *idx),
+            Loc::Stack(idx) => ir.load_stack(reg(dst), *idx),
+            Loc::Mem(idx) => ir.load_mem(reg(dst), *idx),
+            Loc::Param(idx) => ir.load_param(reg(dst), *idx),
         };
 
         dst
@@ -280,34 +338,33 @@ impl Node {
     fn compile_var(&self, ir: &mut dyn Generator, base: u8, pool: &mut Vec<u8>) -> Result<u8> {
         if let Node::Var { sym, status, .. } = &self {
             let mut sym = sym.borrow_mut();
-            let home = ir.first_shadow() + base;
 
             let dst = match status {
                 VarStatus::First => {
                     sym.reg = pool.pop();
                     // if no pool register is available, just use the standard designated register (home)
-                    Self::load_var(ir, sym.reg.unwrap_or(home), &sym.loc)
+                    Self::load_var(ir, sym.reg.unwrap_or(base), &sym.loc)
                 }
                 VarStatus::Mid => {
                     // if no pool register is available, just use the standard designated register (home)
                     // note that this means reloading the variable at each use
                     sym.reg
-                        .unwrap_or_else(|| Self::load_var(ir, home, &sym.loc))
+                        .unwrap_or_else(|| Self::load_var(ir, base, &sym.loc))
                 }
                 VarStatus::Last => {
                     if let Some(r) = sym.reg {
-                        ir.fmov(home, r);
+                        ir.fmov(reg(base), reg(r));
                         pool.push(r);
                         sym.reg = None;
-                        home
+                        base
                     } else {
-                        Self::load_var(ir, home, &sym.loc)
+                        Self::load_var(ir, base, &sym.loc)
                     }
                 }
                 VarStatus::Singular | VarStatus::Unknown => {
                     // if a variable is Singular, i.e., is used only once, don't
                     // bother with caching
-                    Self::load_var(ir, home, &sym.loc)
+                    Self::load_var(ir, base, &sym.loc)
                 }
             };
 
@@ -319,29 +376,25 @@ impl Node {
 
     fn compile_unary(&self, ir: &mut dyn Generator, base: u8, pool: &mut Vec<u8>) -> Result<u8> {
         if let Node::Unary { op, arg, power, .. } = self {
-            let mut dst = ir.first_shadow() + base + self.ershov_number() - 1;
+            let dst = base + self.ershov_number() - 1;
             let r = arg.compile(ir, base, pool)?;
 
             match op.as_str() {
-                "neg" => ir.neg(dst, r),
-                "not" => ir.not(dst, r),
-                "abs" => ir.abs(dst, r),
-                "root" => ir.root(dst, r),
-                "square" => ir.square(dst, r),
-                "cube" => ir.cube(dst, r),
-                "recip" => ir.recip(dst, r),
-                "round" => ir.round(dst, r),
-                "floor" => ir.floor(dst, r),
-                "ceiling" => ir.ceiling(dst, r),
-                "trunc" => ir.trunc(dst, r),
-                "_powi_" => ir.powi(dst, r, *power),
-                "_call_" => {
-                    if r != 0 {
-                        ir.fmov(0, r);
-                    };
-                    dst = 0;
-                }
-                _ => return Err(anyhow!("unary operator is not recognized")),
+                "neg" => ir.neg(reg(dst), reg(r)),
+                "not" => ir.not(reg(dst), reg(r)),
+                "abs" => ir.abs(reg(dst), reg(r)),
+                "root" => ir.root(reg(dst), reg(r)),
+                "square" => ir.square(reg(dst), reg(r)),
+                "cube" => ir.cube(reg(dst), reg(r)),
+                "recip" => ir.recip(reg(dst), reg(r)),
+                "round" => ir.round(reg(dst), reg(r)),
+                "floor" => ir.floor(reg(dst), reg(r)),
+                "ceiling" => ir.ceiling(reg(dst), reg(r)),
+                "trunc" => ir.trunc(reg(dst), reg(r)),
+                "frac" => ir.frac(reg(dst), reg(r)),
+                "_powi_" => ir.powi(reg(dst), reg(r), *power),
+                "_call_" => ir.setup_call_unary(reg(r)),
+                _ => return Err(anyhow!("unary operator {:?} is not recognized", op)),
             };
 
             Ok(dst)
@@ -362,25 +415,25 @@ impl Node {
             let (dst, l, r) = self.alloc(ir, base, left, right, pool)?;
 
             match op.as_str() {
-                "plus" => ir.plus(dst, l, r),
-                "minus" => ir.minus(dst, l, r),
-                "times" => ir.times(dst, l, r),
-                "divide" => ir.divide(dst, l, r),
-                "rem" => ir.fmod(dst, l, r),
-                "gt" => ir.gt(dst, l, r),
-                "geq" => ir.geq(dst, l, r),
-                "lt" => ir.lt(dst, l, r),
-                "leq" => ir.leq(dst, l, r),
-                "eq" => ir.eq(dst, l, r),
-                "neq" => ir.neq(dst, l, r),
-                "and" => ir.and(dst, l, r),
-                "or" => ir.or(dst, l, r),
-                "xor" => ir.xor(dst, l, r),
-                "select_if" => ir.select_if(dst, l, r),
-                "select_else" => ir.select_else(dst, l, r),
-                "_powi_mod_" => ir.powi_mod(dst, l, *power, r),
-                "_call_" => Self::call(ir, l, r),
-                _ => return Err(anyhow!("binary operator is not recognized")),
+                "plus" => ir.plus(reg(dst), reg(l), reg(r)),
+                "minus" => ir.minus(reg(dst), reg(l), reg(r)),
+                "times" => ir.times(reg(dst), reg(l), reg(r)),
+                "divide" => ir.divide(reg(dst), reg(l), reg(r)),
+                "rem" => ir.fmod(reg(dst), reg(l), reg(r)),
+                "gt" => ir.gt(reg(dst), reg(l), reg(r)),
+                "geq" => ir.geq(reg(dst), reg(l), reg(r)),
+                "lt" => ir.lt(reg(dst), reg(l), reg(r)),
+                "leq" => ir.leq(reg(dst), reg(l), reg(r)),
+                "eq" => ir.eq(reg(dst), reg(l), reg(r)),
+                "neq" => ir.neq(reg(dst), reg(l), reg(r)),
+                "and" => ir.and(reg(dst), reg(l), reg(r)),
+                "or" => ir.or(reg(dst), reg(l), reg(r)),
+                "xor" => ir.xor(reg(dst), reg(l), reg(r)),
+                "select_if" => ir.select_if(reg(dst), reg(l), reg(r)),
+                "select_else" => ir.select_else(reg(dst), reg(l), reg(r)),
+                "_powi_mod_" => ir.powi_mod(reg(dst), reg(l), *power, reg(r)),
+                "_call_" => ir.setup_call_binary(reg(l), reg(r)),
+                _ => return Err(anyhow!("binary operator {:?} is not recognized", op)),
             };
 
             Ok(dst)
@@ -399,12 +452,12 @@ impl Node {
     ) -> Result<(u8, u8, u8)> {
         let el = left.ershov_number();
         let er = right.ershov_number();
-        let dst = ir.first_shadow() + base + self.ershov_number() - 1;
+        let dst = base + self.ershov_number() - 1;
 
         let l;
         let r;
 
-        if dst < 16 {
+        if dst < COUNT_SCRATCH {
             if el == er {
                 l = left.compile(ir, base + 1, pool)?;
                 r = right.compile(ir, base, pool)?;
@@ -424,24 +477,6 @@ impl Node {
         Ok((dst, l, r))
     }
 
-    fn call(ir: &mut dyn Generator, l: u8, r: u8) {
-        if l == 1 && r == 0 {
-            ir.fxchg(1, 0);
-        } else if r == 0 {
-            ir.fmov(1, 0);
-            if l != 0 {
-                ir.fmov(0, l);
-            }
-        } else {
-            if l != 0 {
-                ir.fmov(0, l);
-            }
-            if r != 0 {
-                ir.fmov(1, r);
-            }
-        }
-    }
-
     pub fn is_const(&self, val_: f64) -> bool {
         if let Node::Const { val, .. } = self {
             return *val == val_;
@@ -525,6 +560,7 @@ impl Eval for Node {
                     "floor" => x.floor(),
                     "ceiling" => x.ceil(),
                     "trunc" => x.trunc(),
+                    "frac" => x.fract(),
                     "_powi_" => x.powi(*power),
                     "_call_" => {
                         stack[0] = x;
@@ -563,6 +599,17 @@ impl Eval for Node {
                     "xor" => f64::from_bits(x.to_bits() ^ y.to_bits()),
                     "select_if" => f64::from_bits(x.to_bits() & y.to_bits()),
                     "select_else" => f64::from_bits(!x.to_bits() & y.to_bits()),
+                    "min" => f64::min(x, y),
+                    "max" => f64::max(x, y),
+                    "heaviside" => {
+                        if x == 0.0 {
+                            y
+                        } else if x < 0.0 {
+                            0.0
+                        } else {
+                            1.0
+                        }
+                    }
                     "_call_" => {
                         stack[0] = x;
                         stack[1] = y;
@@ -576,3 +623,17 @@ impl Eval for Node {
         }
     }
 }
+
+impl fmt::Debug for Node {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        match self {
+            Node::Void => write!(f, "void"),
+            Node::Const { val, .. } => write!(f, "const {}", val),
+            Node::Var { sym, .. } => write!(f, "var {:?}", sym.borrow()),
+            Node::Unary { op, arg, .. } => write!(f, "{}({:?})", op, arg),
+            Node::Binary {
+                op, left, right, ..
+            } => write!(f, "{}({:?}, {:?})", op, left, right),
+        }
+    }
+}
